% Exemple de code pour SLAM EKF
% (c) 2012 Maxime Latulippe et Philippe Giguere
% Cours GLO-4001/GLO-7021 Introduction a la robotique mobile
% Universite Laval, Quebec, Qc, Canada

clearvars
figure(1)

MakeAvi     = false;  % Pour creer un film avi. Normalement false.
DriveManual = true;   % Pour selectionner les waypoints manuellement
ShowCovariance = true; % Pour montrer la matrice de covariance
Nproche = 0; % Pour s'assurer de toujours avoir quelques points autour de la position de depart

% Ceci est fait pour qu'un instructeur puisse creer un film montrant
% l'evolution du systeme en fonction du temps
if (MakeAvi)
    aviobj = avifile('toto7.avi')
    aviobj.Fps = 5;
    aviobj.KeyFramePerSec = 5;
end

% parametres du systeme
max_range = 2;     % Portee du capteur
map_width = 20;    % Largeur de la carte
map_height = 20;   % Hauteur de la carte
N = 3;             % Nombre de points de reperes sur la carte
dt = 1;            % Intervalle de temps
Speed = 0.5;       % Vitesse lineaire du robot


% Initialisation du robot
robotPos = [0 0 0]';
if ~exist('EKFSLAM_started')
    X = [0 0 0 zeros(1,3*N)]';
    P = zeros(3*N+3); % Matrice de covariances
    P(4:3*N+3,4:3*N+3) = eye(3*N).*999999999;
    seen = [];
    EKFSLAM_started = true;
end

% Positionnement aleatoire des reperes dans l'environnement
x = map_width*rand(1,N) - map_height/2;
y = map_height*rand(1,N) - map_height/2;
if (Nproche~=0)
    x(1:Nproche) = 4*rand(1,Nproche);
    y(1:Nproche) = 4*rand(1,Nproche)-2;
end
features = [x; y];

% bruits
sV = 0.05;
sW = 0.1*pi/180;
sRange = 0.2;
sAngle = 10.0*pi/180;

% Matrices de bruits
R = [0.03 0 0; 0 0.03 0; 0 0 0.003]; % Matrice de covariance sur les deplacements
Q = [0.3 0 0; 0 1.0 0; 0 0 0.00001]; % matrice covariance bruits sur range, angle et signature des features

% Premier trace de la carte
clf
axis([-map_height/2 map_height/2 -map_height/2 map_height/2])
hold on
plot(features(1,:), features(2,:), '.k')
plot(robotPos(1), robotPos(2), 'bo')

IterationNum = 1;  % Iteration ici signifie waypoint

Trajectory = robotPos(1:2);

while (IterationNum<1000) 
    display(sprintf('Iteration %d',IterationNum));
    if (DriveManual)
        display('Cliquez dans l image pour indiquer prochain waypoint');
        [Xtarget Ytarget] = ginput(1);
    else
        if (rand<0.07)
            % Once in a while go to home base, to reduce uncertainty
            Xtarget = 0;
            Yyarget = 0;
        else
            Xtarget = 20*rand-10;
            Ytarget = 20*rand-10;
        end
    end
    Trajectory(:,IterationNum+1) = [Xtarget; Ytarget];
    StepSize = 0.1;
    U = [];
    Straight = [];
    DesiredHeading = atan2(Ytarget-robotPos(2),Xtarget-robotPos(1));
    DeltaHeading = wrapToPi(DesiredHeading-robotPos(3));
    NRotStep = ceil(abs(DeltaHeading)/0.4);
    U(1,1:NRotStep) = 0.0;
    U(2,1:NRotStep) = DeltaHeading/NRotStep; % To make sure we have bang-on rotation;
    
    Distance = sqrt((Xtarget-robotPos(1))^2+(Ytarget-robotPos(2))^2);
    NStraight = ceil(Distance/Speed);
    Straight(1,1:NStraight) = Speed;
    Straight(2,1:NStraight) = 0.001; % Ne peut pas etre zero, sinon jacobienne kaput
    
    U = cat(2,U,Straight);

    for cmd = 1:size(U,2)
        if (cmd>80)
            display('cas probleme');
        end
        % deplacecment du robot
        v = U(1,cmd) + sV*randn;
        w = U(2,cmd) + sW*randn;
        robotPos(1) = robotPos(1) + -v/w*sin(robotPos(3)) + v/w*sin(robotPos(3)+w*dt);
        robotPos(2) = robotPos(2) + v/w*cos(robotPos(3)) - v/w*cos(robotPos(3)+w*dt);
        robotPos(3) = robotPos(3) + w*dt;

        % mesures prises par le robot
        relativePos = features - repmat(robotPos(1:2,:),1,N);
        ranges = sqrt(sum(relativePos.^2,1));
        % conserve uniquement les features visibles selon portee du scanner
        visible = find(ranges <= max_range);
        measures = relativePos(:,visible);
        ranges = ranges(visible);
        angles = atan2(measures(2,:), measures(1,:)) - robotPos(3);
        %display('angles avant bruit')
        %angles
        % ajout de bruit sur les mesures
        ranges = ranges + sRange*randn(1,numel(ranges));
        angles = angles + sAngle*randn(1,numel(angles));
        %display('angles APRES bruit')
        % angles

        % commande executee
        u = U(:,cmd);

        EKFSLAM;  % utilise les valeurs de ranges, angles et u
        %axis([-5 25 -20 20]);
        %axis equal
        
        % L'impression des graphiques se fait a des moments precis, pour
        % reduire le workload
        if  (DriveManual) || (IterationNum < 4) || ((cmd==1) && (IterationNum<40) ) || ( (cmd==1) && (mod(IterationNum,20) == 0) && (IterationNum>50) )
            clf
            axis([-map_height/2 map_height/2 -map_height/2 map_height/2])
            hold on
            hl = line(Trajectory(1,:),Trajectory(2,:));
            set(hl,'Color',[0.7 0.7 0.7]);
            plot(features(1,:), features(2,:), '.k');
            plot(robotPos(1), robotPos(2), 'bo');
            line([robotPos(1) robotPos(1)+0.5*cos(robotPos(3))], [robotPos(2) robotPos(2)+0.5*sin(robotPos(3))], 'color', 'b');
            plotCircle(robotPos(1), robotPos(2), max_range, 'k');
            
            plot(X(1), X(2), 'go');
            line([X(1) X(1)+0.5*cos(X(3))], [X(2) X(2)+0.5*sin(X(3))], 'color', 'g');
            C = P(1:2,1:2);
            herr = plotError([X(1) X(2)], C, 'g', 50);
            set(herr,'LineWidth',2);

            for i = 1:N
                if ~isempty(find(seen == i, 1))
                    x = X(3*i+1);
                    y = X(3*i+2);
                    C = P(3*i+1:3*i+2, 3*i+1:3*i+2);
                    plot(x, y, 'rx');
                    herr = plotError([x y], C, 'r', 50);
                    set(herr,'LineWidth',2);
                    htext = text(x+0.13,y,sprintf('%d',i));
                    set(htext,'Color','r');
                end
            end
            title(sprintf('Trajectoire rectiligne %d',IterationNum),'FontSize',12);
            set(gca,'FontSize',12);

            % Pour montrer matrice de covariance
            if (ShowCovariance)
                figure(2)
                if (N~=3)
                    error('Pour cet exemple, on a besoin de N=3 landmarks');
                end
                clf
                imagesc(log(abs(P)));
                set(gca,'XTick',1:size(P,1))
                set(gca,'YTick',1:size(P,1))
                Labels = {'x' 'y' 'a' 'm1x' 'm1y' 'm1s' 'm2x' 'm2y' 'm2s' 'm3x' 'm3y' 'm3s'};
                set(gca,'XTickLabel',Labels);
                set(gca,'YTickLabel',Labels);
                set(gca,'XAxisLocation','top')
                title('Matrice de covariance P');
                hl = line([0.5 3.5 3.5 0.5 0.5],[0.5 0.5 3.5 3.5 0.5]);
                set(hl,'LineWidth',3','Color','k');
                figure(1)
            end
            drawnow();
            if (MakeAvi)
                % The figure must be in the first screen with dual monitor
                F = getframe(gcf);
                aviobj = addframe(aviobj,F);
            end
        end
    end
    IterationNum = IterationNum + 1;
end


if (MakeAvi)
   aviobj = close(aviobj)
end